/*
  Boite histoires pour enfants
  Plays MP3 file from microSD card
  Uses MAX98357 I2S Amplifier Module
  Uses ESP32-audioI2S Library - https://github.com/schreibfaul1/ESP32-audioI2S
  * 
*/

#include <WiFi.h>
#include "ESPTelnet.h"        
#include "ESP32FtpServer.h"  

// Include required libraries
#include "Arduino.h"
#include <ArduinoOTA.h>
#include "Audio.h"
#include "SD.h"
#include "FS.h"
#include "Touch.h"
#include <Preferences.h>
#include <ESPmDNS.h>
#include "esp_wps.h"

/* ------------------------------------------------- */


ESPTelnet telnet;
FtpServer ftpSrv;
IPAddress ip;
uint16_t  portTelnet = 23;
int TelnetConnected = 0;
uint32_t TimeStampStartWifi;
uint32_t TimeStampEndWifi;
uint32_t TimeStampConnectWifi;
bool WifiON;
bool FirstStart = false;

int FlagWPSActif = 0;
int FlagWPSDone = 0;
int ForceWPS;
uint32_t NextStepWPS;
uint32_t LastMessageWifi;
uint32_t NextChangeLed;
int NextLEDState;

int TryConnect;
int mDNSStarted = 0;


 void SaveDataNVS();
 void SetStoryList();
 int CheckBatLevel();

// microSD Card Reader connections
#define SD_CS          4
#define SPI_MOSI      23 
#define SPI_MISO      19
#define SPI_SCK       18
 
// I2S Connections
//  Warning, on the first board the pin #26 which is the default for BCLK is defective. Do NOT use that pin.
#define I2S_DOUT      25
#define I2S_BCLK      12
#define I2S_LRC       22
#define I2S_SD        33

//  Touch PADs 
//  Using IO0 can lead to strange effects, do not use this special PIN for touch PAD

#define PAD_OK_Num    2
#define PAD_FWD_Num   14
#define PAD_BACK_Num  15
//  Default value for threshold, will be updated (calibration)
#define THRESHOLD_OK    113
#define THRESHOLD_FWD   104
#define THRESHOLD_BACK  115

#define PIN_VBAT  35
#define PIN_VBUS  36

#define THRESHOLD_VBUS 2500
#define PIN_LED   27


Touch Pad_OK(PAD_OK_Num, THRESHOLD_OK);
Touch Pad_FWD(PAD_FWD_Num, THRESHOLD_FWD);
Touch Pad_BACK(PAD_BACK_Num, THRESHOLD_BACK);

bool FlagCalibration = false;
bool CalibrationOK = false;
bool CalibrationFWD = false;
bool CalibrationBACK = false;


#define KEY_OK      1
#define KEY_FWD     2
#define KEY_BACK    4

//  Define debounce delay (in µS)
//  2 different delays for High to Low (OK) or Low to High
#define DebounceDelay_HL     100000       //  100ms
#define DebounceDelay_LH     10000        //  10ms


 // Create Audio object
Audio audio;

#define MAX_VOLUME 22

int nHistoires = 0;
bool PlayingTitle = false;
bool PlayingStory = false;
bool ModeChangeVol = false;
int CurrentStory;
uint32_t EndWaitForOK;
uint32_t TS_NextStory;

int CounterCritical;      //  After 3 times with battery status critical, shutdown the ESP

enum STATUS_STORY {
  WAIT_FOR_START,
  READING_TITLE,
  WAIT_FOR_OK,
  READING_STORY,
  READING_PAUSE,
  WAIT_FOR_NEXT_STORY,
  STORY_END,
  STATE_ERROR,
  CHANGE_VOLUME_START,
  CHANGE_VOLUME_NEW,
  CHANGE_VOLUME_WAIT,
  CHANGE_VOLUME_1,
};
STATUS_STORY CurrentState = WAIT_FOR_START;

#define MAX_BACK_STORY  4

int IndexRecentStories[MAX_BACK_STORY+1];

Preferences preferences;        //  Store values in non volatile storage
int CurrentVolume;
bool ModeContinu;
bool isCharging;
uint32_t LastSeenUSBOff;
uint32_t WifiTimeOut;


bool isConnected() {
  return (WiFi.status() == WL_CONNECTED);
}

#define ESP_WPS_MODE WPS_TYPE_PBC

void wpsStart() {
  esp_wps_config_t config;
  memset(&config, 0, sizeof(esp_wps_config_t));
  //Same as config = WPS_CONFIG_INIT_DEFAULT(ESP_WPS_MODE);
  config.wps_type = ESP_WPS_MODE;
  strcpy(config.factory_info.manufacturer, "TH");
  strcpy(config.factory_info.model_number, CONFIG_IDF_TARGET);
  strcpy(config.factory_info.model_name, "LOLIN_D32_PRO");
  strcpy(config.factory_info.device_name, "BOITE_HISTOIRE");
  esp_err_t err = esp_wifi_wps_enable(&config);
  if (err != ESP_OK) {
    Serial.printf("WPS Enable Failed: 0x%x: %s\n", err, esp_err_to_name(err));
    return;
  }

  err = esp_wifi_wps_start(0);
  if (err != ESP_OK) {
    Serial.printf("WPS Start Failed: 0x%x: %s\n", err, esp_err_to_name(err));
  }
}

void wpsStop() {
  esp_err_t err = esp_wifi_wps_disable();
  if (err != ESP_OK) {
    Serial.printf("WPS Disable Failed: 0x%x: %s\n", err, esp_err_to_name(err));
  }
}

String wpspin2string(uint8_t a[]) {
  char wps_pin[9];
  for (int i = 0; i < 8; i++) {
    wps_pin[i] = a[i];
  }
  wps_pin[8] = '\0';
  return (String)wps_pin;
}

// WARNING: WiFiEvent is called from a separate FreeRTOS task (thread)!
void WiFiEvent(WiFiEvent_t event, arduino_event_info_t info) {
  switch (event) {
    case ARDUINO_EVENT_WIFI_STA_START: 
      Serial.println("Station Mode Started"); 
      break;
    case ARDUINO_EVENT_WIFI_STA_GOT_IP:
      Serial.println("Connected to :" + String(WiFi.SSID()));
      Serial.print("Got IP: ");
      Serial.println(WiFi.localIP());
      break;
    case ARDUINO_EVENT_WIFI_STA_DISCONNECTED:
      Serial.println("Disconnected from station, attempting reconnection");
      WiFi.reconnect();
      break;
    case ARDUINO_EVENT_WPS_ER_SUCCESS:
      Serial.println("WPS Successful, stopping WPS and connecting to: " + String(WiFi.SSID()));
      wpsStop();
      delay(10);
      FlagWPSActif = 0;
      ForceWPS = 0;
      FlagWPSDone = 1;
      SaveDataNVS();
      WiFi.begin();
      break;
    case ARDUINO_EVENT_WPS_ER_FAILED:
      Serial.println("WPS Failed, retrying");
      wpsStop();
      wpsStart();
      break;
    case ARDUINO_EVENT_WPS_ER_TIMEOUT:
      Serial.println("WPS Timeout, retrying");
      wpsStop();
      wpsStart();
      break;
    case ARDUINO_EVENT_WPS_ER_PIN: Serial.println("WPS_PIN = " + wpspin2string(info.wps_er_pin.pin_code)); break;
    default:                       break;
  }
}

/* ------------------------------------------------- */


//  Callback functions for FTP events

void InfoFTPConnect(String ip)
{
  Serial.printf("FTP connect IP=%s is now connected\n", ip);
  telnet.printf("FTP connect IP=%s is now connected\n", ip);
  TimeStampEndWifi = millis() + WifiTimeOut;
}

void InfoFTPDisconnect(String ip)
{
  Serial.printf("FTP disconnect IP=%s is now disconnected\n", ip);
  telnet.printf("FTP disconnect IP=%s is now disconnected\n", ip);
}

// (optional) callback functions for telnet events
void onTelnetConnect(String ip) {
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" connected");
  
  telnet.println("\nWelcome " + telnet.getIP());
  telnet.println("(Use ^] + q  to disconnect.)");
  TelnetConnected = true;
  TimeStampEndWifi = millis() + WifiTimeOut;   //  Wifi off after Timeout (10mn default)
}

void onTelnetDisconnect(String ip) {
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" disconnected");
  TelnetConnected = false;
}

void onTelnetReconnect(String ip) {
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" reconnected");
  TelnetConnected = true;
  
}

void onTelnetConnectionAttempt(String ip) {
  Serial.print("- Telnet: ");
  Serial.print(ip);
  Serial.println(" tried to connected");
}

void onTelnetInput(String str) {
  //Change telnet end timer when a command is received.
  TimeStampEndWifi = millis() + WifiTimeOut;   //  Wifi off after Timeout (10mn default)
  // checks for a certain command
  if (str == "ping") {
    telnet.println("> pong");
    Serial.println("- Telnet: pong");
  // disconnect the client
  } 
  else if (str == "bye") 
  {
    telnet.println("> disconnecting you...");
    telnet.disconnectClient();
    TimeStampEndWifi = millis() + 30000;    //  Stop Wifi after 30s after disconnect
  } 
  else if ( str== "cal")
  {
    telnet.println("Start calibration touch PADs");
    Serial.println("Start calibration touch PADs");
    FlagCalibration = true;
    CalibrationOK = true;
    CalibrationBACK = false;
    CalibrationFWD = false;
  }
  else if ( str == "dns")
  {
    telnet.printf("mDNS status = %d\n", mDNSStarted);
    Serial.printf("mDNS status = %d\n", mDNSStarted);
  }
  else if ( str == "usb" )
  {
    int curVBUS = analogRead(PIN_VBUS);
    if ( isCharging )
    {
      telnet.printf("USB is on since %.1f s, VBUS=%.1f\n", (millis() - LastSeenUSBOff)/1000.0, curVBUS*6.6/4096);
      Serial.printf("USB is on since %.1f s, VBUS=%.1f\n", (millis() - LastSeenUSBOff)/1000.0, curVBUS*6.6/4096);
    }
    else
    {
      telnet.printf("USB is off, VBUS = %1.f\n", curVBUS*6.6/4096);
      Serial.printf("USB is off, VBUS = %1.f\n", curVBUS*6.6/4096);
    }
  }
  else if ( str == "vol" )
  {
    telnet.printf("Current volume is %d\n", CurrentVolume);
    Serial.printf("Current volume is %d\n", CurrentVolume);
  }
  else if ( str == "forceWPS")
  {
    ForceWPS = 1;
    SaveDataNVS();
    telnet.println("ForceWPS is now ON, Restart ESP in 2s...");
    Serial.println("ForceWPS is now ON, Restart ESP in 2s...");
    for ( int i = 0; i < 10; i++ )
    {
      telnet.loop();
      delay(100);
    }
    telnet.disconnectClient();
    for ( int i = 0; i < 10; i++ )
    {
      telnet.loop();
      delay(100);
    }
    ESP.restart();
  }
  else if ( str == "clearnvs")
  {
    preferences.clear();
    telnet.println("All data on nvs ois now cleared");
    Serial.println("All data on nvs ois now cleared");
    telnet.println("Nvs cleared, Restart ESP in 2s...");
    Serial.println("Nvs cleared, Restart ESP in 2s...");
    for ( int i = 0; i < 10; i++ )
    {
      telnet.loop();
      delay(100);
    }
    telnet.disconnectClient();
    for ( int i = 0; i < 10; i++ )
    {
      telnet.loop();
      delay(100);
    }
    ESP.restart();
  }
  else if ( str[0] == 'c' && str[1] == 'o' && str[2] == 'n' &&  str[3] == 't' && str[4] == ' ')
  {
    int ValContinu = 0;
    if ( sscanf(str.c_str()+5, "%d" , &ValContinu) == 1 )
    {
      if ( ValContinu == 0 || ValContinu == 1 )
      {
        ModeContinu = ValContinu;
        telnet.printf("Mode continu is %d\n", ModeContinu);
        Serial.printf("Mode continu is %d\n", ModeContinu);
        SaveDataNVS();
      }
      else
      {
        Serial.printf("Error, mode continu should be 0 or 1, asked %d\n", ValContinu);
        telnet.printf("Error, mode continu should be 0 or 1, asked %d\n", ValContinu);
      }
    }
  }
  else if ( str == "pad" )
  {
    telnet.printf("Current pad thresholds are Pad_OK:%d  Pad_FWD:%d, Pad_BACK:%d\n", 
        Pad_OK.getThreshold(), Pad_FWD.getThreshold(), Pad_BACK.getThreshold());
    Serial.printf("Current pad thresholds are Pad_OK:%d  Pad_FWD:%d, Pad_BACK:%d\n", 
        Pad_OK.getThreshold(), Pad_FWD.getThreshold(), Pad_BACK.getThreshold());
  }
  else if ( str == "bat" )
  {
    CheckBatLevel();
  }
  else if ( str == "stop" )
  {
    if ( audio.isRunning() ) audio.stopSong();
    CurrentState = STORY_END;
    PlayingStory = false;
    PlayingTitle = false;
  }
  else if (str[0] == 'p' && str[1] == 'a' && str[2] == 'd' &&  str[3] == ' ' )
  {
    int ValPAD = 0;
    if ( sscanf(str.c_str()+4, "OK %d" , &ValPAD) == 1 )
    {
      Serial.printf("Set Threshold for PAD OK to %d\n", ValPAD);
      telnet.printf("Set Threshold for PAD OK to %d\n", ValPAD);
      Pad_OK.setThreshold(ValPAD);
      SaveDataNVS();
    }
    if ( sscanf(str.c_str()+4, "FWD %d" , &ValPAD) == 1 )
    {
      Serial.printf("Set Threshold for PAD FWD to %d\n", ValPAD);
      telnet.printf("Set Threshold for PAD FWD to %d\n", ValPAD);
      Pad_FWD.setThreshold(ValPAD);
      SaveDataNVS();
    }
    if ( sscanf(str.c_str()+4, "BACK %d" , &ValPAD) == 1 )
    {
      Serial.printf("Set Threshold for PAD BACK to %d\n", ValPAD);
      telnet.printf("Set Threshold for PAD BACK to %d\n", ValPAD);
      Pad_BACK.setThreshold(ValPAD);
      SaveDataNVS();
    }
  }
  else if (str[0] == 'v' && str[1] == 'o' && str[2] == 'l' &&  str[3] == ' ' )
  {
    int ValVolume = 0;
    if ( sscanf(str.c_str()+4, "%d" , &ValVolume) == 1 )
    {
      if ( ValVolume > 0 && ValVolume <= 22 )
      {
        Serial.printf("Set Volume to %d\n", ValVolume);
        telnet.printf("Set Volume to %d\n", ValVolume);
        audio.setVolume(ValVolume);
        CurrentVolume = ValVolume;
        SaveDataNVS();
      }
      else
      {
        Serial.printf("Error, volume should be set between 1 and 22, asked %d\n", ValVolume);
        telnet.printf("Error, volume should be set between 1 and 22, asked %d\n", ValVolume);
      }
    }
  }
  else if ( str[0] == 'p' && str[1] == 'l' && str[2] == 'a' && str[3] == 'y' && str[4] == ' ')
  {
    int NumStory = 0;
    char Buffer[32];
    if ( sscanf(str.c_str()+5, "%d" , &NumStory) == 1 )
    {
      if ( NumStory > 0 && NumStory <=  nHistoires )
      {
        if (audio.isRunning() ) audio.stopSong();
        Serial.printf("Set Story to %d\n", NumStory);
        telnet.printf("Set Story to %d\n", NumStory);
        CurrentStory = NumStory;
        CurrentState = READING_STORY;
        PlayingTitle = false;
        SetStoryList();
        snprintf(Buffer, 32, "/Stories/H%03d.mp3", CurrentStory);
        audio.connecttoFS(SD, Buffer);
        PlayingStory = true;
      }
      else
      {
        Serial.printf("Error, Story should be set between 1 and %d, asked %d\n", nHistoires, NumStory);
        telnet.printf("Error, Story should be set between 1 and %d, asked %d\n", nHistoires, NumStory);
      }
    }
  }
  else if ( str == "timeout" )
  {
    Serial.printf("TimeOut Wifi %ld s, shutdown in %ds\n", WifiTimeOut/1000, (TimeStampEndWifi - millis())/1000);
    telnet.printf("TimeOut Wifi %ld s, shutdown in %ds\n", WifiTimeOut/1000, (TimeStampEndWifi - millis())/1000);
  }
  else if ( str[0] == 't' && str[1] == 'i' && str[2] == 'm' && str[3] == 'e' && str[4]== 'o' && str[5] == 'u' && str[6] == 't' && str[7] == ' ')
  {
    //  Set Wifi Time out
    int32_t Val;
    if ( sscanf(str.c_str()+8, "%d" , &Val) == 1 )
    {
      if ( Val > 0 && Val < 1000000 )
      {
        Val *= 1000;    //  Pour avoir des ms...
        WifiTimeOut = Val;
        TimeStampEndWifi = millis() + Val;
        Serial.printf("Set Wifi Timeout to %d s\n", Val/1000);
        telnet.printf("Set Wifi Timeout to %d s\n", Val/1000);
      }
    }

  }
  else if ( str== "rst")
  {
    telnet.println("Restart ESP in 2s...");
    Serial.println("Restart ESP in 2s...");
    for ( int i = 0; i < 10; i++ )
    {
      telnet.loop();
      delay(100);
    }
    telnet.disconnectClient();
    for ( int i = 0; i < 10; i++ )
    {
      telnet.loop();
      delay(100);
    }
    ESP.restart();
  }
}

/* ------------------------------------------------- */


void errorMsg(String error, bool restart = true) {
  Serial.println(error);
  if (restart) {
    Serial.println("Rebooting now...");
    delay(2000);
    ESP.restart();
    delay(2000);
  }
}

//  Lance Wifi mais n'attend PAS le résultat de connexion

void StartWifiConnection()
{
  TimeStampStartWifi = millis();
  TimeStampEndWifi = TimeStampStartWifi + 120000;     //  Coupe le Wifi si pas connecté après 2mn
  WiFi.mode(WIFI_STA);
  WiFi.setHostname("Histoire");
  WiFi.begin();

}

/* ------------------------------------------------- */

void setupTelnet() {  
  // passing on functions for various telnet events
  telnet.onConnect(onTelnetConnect);
  telnet.onConnectionAttempt(onTelnetConnectionAttempt);
  telnet.onReconnect(onTelnetReconnect);
  telnet.onDisconnect(onTelnetDisconnect);
  telnet.onInputReceived(onTelnetInput);

  Serial.print("- Telnet: ");
  if (telnet.begin(portTelnet)) {
    Serial.println("running");
  } else {
    Serial.println("error.");
    errorMsg("Will reboot...");
  }
  MDNS.addService("telnet", "tcp", 23);
}



#define THRESHOLD_BAT_0     3.1      //  Battery level 1 --> nearly 0%  3.1V
#define THRESHOLD_BAT_1     3.5     //  Battery level 1 --> 20%  : 3.5V
#define THRESHOLD_BAT_FULL  4.15    //  Battery full : 4.15V

#define DELAY_BAT_CRITICAL    1000*60*1   //  Check every mn
#define DELAY_BAT_LOW         1000*60*5   //  Check every 5 mn
#define DELAY_BAT_OK          1000*60*5   //  Check every 5 mn

enum BAT_Status { STATUS_OK=0, STATUS_LOW, STATUS_CRITICAL, STATUS_FULL };

int LastBatStatus;
unsigned long NextCheckBat;
const float correction_adc = 1.11;

int CheckBatLevel()
{
char Buffer[128];
int nMeasure = 16;
unsigned long SumVBat = 0;
float VBat;
int Min, Max;

  Max = 0;
  Min = 4095;
  for (int i = 0; i < nMeasure; i++ )
  {
    unsigned int raw_vbat = analogRead(PIN_VBAT);
    if ( raw_vbat < Min )
      Min = raw_vbat;
    if ( raw_vbat > Max )
      Max = raw_vbat;
    delay(1);
    SumVBat += raw_vbat;
  }
  int iVBat = (SumVBat / nMeasure);
  VBat = (iVBat / 4095.0)*6.6*correction_adc;
  snprintf(Buffer, 126, "Battery Level = %.2f\n", VBat);
  Serial.print(Buffer);
  if ( WifiON )
    telnet.print(Buffer);
  if ( VBat < THRESHOLD_BAT_0 )   //  Critical level
  {
      Serial.println("Battery Status : critical");
      if ( WifiON )
        telnet.println("Battery Status : critical");
      if ( audio.isRunning() ) audio.stopSong();
      delay(500);
      audio.connecttoFS(SD, "/BatCrit.mp3");      //  Sortie message audio dans tous les cas
      delay(2000);                                //  Attente
      NextCheckBat = millis() + DELAY_BAT_CRITICAL;
      LastBatStatus = STATUS_CRITICAL;
      CounterCritical++;
      if ( CounterCritical >= 3 )
      {
        Serial.println("Going to sleep now");
        delay(1000);
        if ( WifiON )
        {
          telnet.println("Going to sleep now");
          for ( int i = 0; i < 10; i++ )
          {
            telnet.loop();
            delay(50);
          }
        }
      
        Serial.flush(); 
        esp_deep_sleep_start();
        Serial.println("This will never be printed");
      }
      return STATUS_CRITICAL;
  }
  if ( VBat < THRESHOLD_BAT_1 && LastBatStatus != STATUS_CRITICAL )   //  Low level
  {
      Serial.println("Battery Status : LOW");
      if ( WifiON )
        telnet.println("Battery Status : LOW");
      if ( !audio.isRunning() ) 
      {
        audio.connecttoFS(SD, "/BatLow.mp3");   //  Sortie message audio si pas en lecture d'histoire
        delay(2000);
      }
      NextCheckBat = millis() + DELAY_BAT_LOW;
      LastBatStatus = STATUS_LOW;
      CounterCritical = 0;
      return STATUS_LOW;
  }
  Serial.println("Battery status : OK");
  NextCheckBat = millis() + DELAY_BAT_OK;
  LastBatStatus = STATUS_OK;
  if ( VBat > THRESHOLD_BAT_FULL )
  {
    LastBatStatus = STATUS_FULL;
  }
  CounterCritical = 0;
  return STATUS_OK;
}

//  Liste le contenu des répartoires de la carte SD
//  Aussi utilisé pour mémoriser le nuombre d'histoires

void printDirectory(File dir, int numTabs, bool FlagStories) {
  int nFileDir = 0;

  while (true) {
    File entry = dir.openNextFile();
    if (!entry) {
      if (numTabs == 0)
        Serial.println("** Done **");
      if ( FlagStories )
      {
        //Serial.printf("Répertoire Stories avec %d fichiers\n", nFileDir);
        nHistoires = nFileDir / 2;
      }
      return;
    }
    nFileDir++;

    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }

    Serial.print(entry.name());
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1, strcasecmp(entry.name(), "Stories") == 0);
    } else {
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }

    entry.close();
  }
}


//  Sauve les préférences dans la mémoire non volatile

void SaveDataNVS()
{

  //  Access data stored in nvs. If does not exist yet open in write mode
  preferences.begin("StoryBox", false);
  
  //  Save measure to retrieve them on next start...
  preferences.putLong("CurrentVolume", CurrentVolume);
  preferences.putLong("RecentStory-M4", IndexRecentStories[0]);
  preferences.putLong("RecentStory-M3", IndexRecentStories[1]);
  preferences.putLong("RecentStory-M2", IndexRecentStories[2]);
  preferences.putLong("RecentStory-M1", IndexRecentStories[3]);
  preferences.putInt("Threshold_OK", Pad_OK.getThreshold());
  preferences.putInt("Threshold_BACK", Pad_BACK.getThreshold());
  preferences.putInt("Threshold_FWD", Pad_FWD.getThreshold());
  preferences.putLong("TimeOut", WifiTimeOut);
  preferences.putInt("ModeContinu", ModeContinu);
  preferences.putInt("ForceWPS", ForceWPS);

  preferences.end();    //  Close NVS
  Serial.printf("Sauve Valeurs Preferences : Vol=%d, ModeContinu=%d, M4=%d, M3=%d, M2=%d, M1=%d, Seuil_OK=%d, Seuil_FWD=%d, Seuil_BACK=%d, WifiTimeOut=%ld, ForceWPS=%d\n", 
      CurrentVolume , ModeContinu, IndexRecentStories[0], IndexRecentStories[1], IndexRecentStories[2], IndexRecentStories[3],
      Pad_OK.getThreshold(), Pad_FWD.getThreshold(), Pad_BACK.getThreshold(), WifiTimeOut/1000, ForceWPS);

}


//  Retourne le numéro de la prochaine histoire
//  Utilise une variable pseudo aléatoire basée sur la ms courante
//  Le numéro d'histoire retourné commence à 1...

int getNextStory()
{
uint32_t randVal = (millis() % nHistoires);
bool EndNext = false;

  while ( 1 )
  {
    int i;
    //  Vérifie que cette histoire n'est PAS dans les récentes
    for ( i = 0; i < MAX_BACK_STORY; i++)
    {
      if ( IndexRecentStories[i] == randVal ) break;      //  Déjà présente ?  
    }
    if ( i >= MAX_BACK_STORY )
      break;
    randVal++;          //  Passe à la suivante
    randVal %= nHistoires;
  }
  IndexRecentStories[MAX_BACK_STORY] = randVal;       //  La met à la fin de la liste
  return randVal + 1;
}

//  Copie l'histoire en cours dans la liste des histoires et réarrange les index

void SetStoryList()
{
int StartIndex = 0;

  //  Tout d'abord cherche si l'histoire en cours est dans la liste
  for (int i = 0; i < MAX_BACK_STORY; i++)
  {
    if ( IndexRecentStories[i] == IndexRecentStories[MAX_BACK_STORY] )
    {
      //  Oui dans la liste, ne va écraser les histoires qu'à partir de celle ci
      StartIndex = i;
    }
  }
  for (int i = StartIndex; i < MAX_BACK_STORY; i++)
  {
    IndexRecentStories[i] = IndexRecentStories[i+1];
  }
  SaveDataNVS();
}

// Retourne le numéro de l'histoire précedente.
//  Retourne -1 si pas possible

int getPrevStory()
{
  Serial.printf("getPrevStory: Looking for story before %d\n", IndexRecentStories[MAX_BACK_STORY]);
  for ( int i = 0; i < MAX_BACK_STORY; i++)
  {
    if ( IndexRecentStories[i] == IndexRecentStories[MAX_BACK_STORY] )
    {
      //  C'est celle ci, recule d'une case si possible
      if ( i > 0 )
      {
        i--;
        IndexRecentStories[MAX_BACK_STORY] = IndexRecentStories[i];
        Serial.printf("getPrevStory: Found at index %d, new list is M1=%d M2=%d M3=%d M4=%d Current=%d\n"
          , i, IndexRecentStories[0], IndexRecentStories[1], IndexRecentStories[2], IndexRecentStories[3] , IndexRecentStories[4]);
        return IndexRecentStories[MAX_BACK_STORY] + 1;
      }
      Serial.printf("Error can't rewind before first story, return -1\n");
      return -1;      //  Erreur
    }
  }
  IndexRecentStories[MAX_BACK_STORY] = IndexRecentStories[MAX_BACK_STORY-1];
  return IndexRecentStories[MAX_BACK_STORY] + 1;
}

//  Code exécuté après redémarrage de l'ESP
//  Initialisation....

void setup() {
int TrySD;
    
    pinMode(BUILTIN_LED, OUTPUT);
    pinMode(PIN_LED, OUTPUT);
    
    digitalWrite(BUILTIN_LED, HIGH);        // Build in LED Off
    digitalWrite(PIN_LED, HIGH);            // LED ON
    // Set microSD Card CS as OUTPUT and set HIGH
    pinMode(SD_CS, OUTPUT);      
    digitalWrite(SD_CS, HIGH); 


    // Start Serial Port
    Serial.begin(115200);
    delay(500);
    Serial.println("\n\nStarting StoryBox v1");

    

    
  WifiON = true;
  digitalWrite(BUILTIN_LED, LOW);     //  Built in LED ON now 
  
    // Initialize SPI bus for microSD Card
    SPI.begin(SPI_SCK, SPI_MISO, SPI_MOSI);
    
  if ( WifiON )
    telnet.print("Checking SD Card...");
    // Start microSD Card
    for ( TrySD = 10; TrySD> 0; TrySD-- )
    {
      int StatusSD = SD.begin(SD_CS);
      if ( StatusSD ) break;
      delay(100);
      digitalWrite(PIN_LED, TrySD & 1);
      digitalWrite(BUILTIN_LED, TrySD & 1);
      Serial.print("Error mounting SD Card, remaining tries "); Serial.print(TrySD); Serial.println();
      telnet.print("Error mounting SD Card, remaining tries "); telnet.print(TrySD); telnet.println();
      
    }
    if( TrySD <= 0 )
    {
      Serial.println("Error accessing microSD card!");
      telnet.println("Error accessing microSD card!");
      for ( int i = 0; i < 100; i++)
      {
        digitalWrite(PIN_LED, i & 1);
        digitalWrite(BUILTIN_LED, i & 1);
        delay(100);
      }
      ESP.restart();
    }
    telnet.println("  OK");
    Serial.println("  OK");
    digitalWrite(PIN_LED, HIGH);
    digitalWrite(BUILTIN_LED, HIGH);

  // Lecture des fichiers de la carte SD

  File root = SD.open("/");

  printDirectory(root, 0, false);
  root.close();
 

  touchSetCycles(0x2000, 0x1000);
  Pad_OK.setName("OK");
  Pad_BACK.setName("Back");
  Pad_FWD.setName("Fwd");

  Pad_OK.setStatPeriod(0);
  Pad_BACK.setStatPeriod(00);
  Pad_FWD.setStatPeriod(1000);
  
  Serial.print("Now start playing\n");
  if ( WifiON ) telnet.print("Now start playing\n");

  Serial.printf("%d histoires sur la carte SD\n", nHistoires);


    //  Retrieve data stored in nvs. If does not exist yet open in write mode
  preferences.begin("StoryBox", false);
  
  //  Default value is 16 medium
  CurrentVolume =   preferences.getLong("CurrentVolume", 16);
  //  Valeurs par défaut 1,2,3,4
  IndexRecentStories[0] = preferences.getLong("RecentStory-M4", 1);
  IndexRecentStories[1] = preferences.getLong("RecentStory-M3", 2);
  IndexRecentStories[2] = preferences.getLong("RecentStory-M2", 3);
  IndexRecentStories[3] = preferences.getLong("RecentStory-M1", 4);
  Pad_OK.setThreshold(preferences.getInt("Threshold_OK", THRESHOLD_OK));
  Pad_FWD.setThreshold(preferences.getInt("Threshold_FWD", THRESHOLD_FWD));
  Pad_BACK.setThreshold(preferences.getInt("Threshold_BACK", THRESHOLD_BACK));
  WifiTimeOut = preferences.getLong("TimeOut", 600000);
  ModeContinu = preferences.getInt("ModeContinu", 100);
  ForceWPS  = preferences.getInt("ForceWPS", 1);
  preferences.end();
  Serial.printf("Valeurs Preferences : Vol=%d, ModeContinu=%d, M4=%d, M3=%d, M2=%d, M1=%d, TimeOut=%ld, ForceWPS=%d, Seuil OK=%d FWD=%d  BACK=%d \n", 
      CurrentVolume , ModeContinu, IndexRecentStories[0], IndexRecentStories[1], IndexRecentStories[2], IndexRecentStories[3], 
      WifiTimeOut, ForceWPS,
      Pad_OK.getThreshold(), Pad_FWD.getThreshold(), Pad_BACK.getThreshold());
  FlagWPSActif = ForceWPS;
  if ( FlagWPSActif )
  {
    Serial.println("ForceWPS actif, appuyez sur le bouton du routeur");
    Serial.print("Starting WPS .");
    WiFi.onEvent(WiFiEvent);  // Will call WiFiEvent() from another thread.
    WiFi.mode(WIFI_MODE_STA);
    wpsStart();      
    NextStepWPS = millis()+500;
    TimeStampEndWifi = millis() + 210000; //  Wifi On for at least 3'30 mn
    FlagWPSActif = 2;
    ForceWPS = 0;
  }
  else
  {
    Serial.print("- Wifi: ");
    StartWifiConnection();  
  }
  if ( ModeContinu > 1 )
  {
    ModeContinu = 1;
    FirstStart = true;
    FlagCalibration = true;
    CalibrationOK = true;
    CalibrationBACK = false;
    CalibrationFWD = false;
  }
    // Setup I2S 
    pinMode(I2S_SD, OUTPUT);
  
    audio.setPinout(I2S_BCLK, I2S_LRC, I2S_DOUT);
    //audio.i2s_mclk_pin_select(1);     //  Avoid MCLK on pin 0/

  CurrentState = WAIT_FOR_START;
    
    // Set Volume
    audio.setVolume(CurrentVolume);
    digitalWrite(I2S_SD, HIGH);

  int CurrentVBUS = analogRead(PIN_VBUS);
  LastSeenUSBOff = millis();
  isCharging = false;
  if ( CurrentVBUS > THRESHOLD_VBUS )
  {
    Serial.printf("USB is powered, PIN_VBUS=%d\n", CurrentVBUS);
  }
  else
  {
    Serial.printf("USB is left open, PIN_VBUS=%d\n", CurrentVBUS);
  }
  //  Check battery level
  //  Playing sound in startup should be the last item, as audio.loop is needed to actually play sound.
  if ( CheckBatLevel() == STATUS_OK )
  {
    // Open music file
    audio.connecttoFS(SD,"/Start.mp3");
  }

  if ( FirstStart )
  {
    Serial.println("First Start, Mode calibration will be active");
    for (int i = 0; i < 5000; i++)
    {
      audio.loop();
      delay(1);
    }
    FirstStart = 0;
  }

}
 
uint32_t lastLoopPrint;

//  Code exécuté régulièrement (boucle principale Arduino).


void loop()
{
char Buffer[32];

  if ( WifiON )
  {
    if ( FlagWPSActif > 0 )
    {
      if ( FlagWPSDone )
      {
        Serial.printf("WPS done! ");
        FlagWPSActif = 0;
      }
      if ( FlagWPSActif == 1) 
      {
        Serial.print("Starting WPS .");
        WiFi.onEvent(WiFiEvent);  // Will call WiFiEvent() from another thread.
        WiFi.mode(WIFI_MODE_STA);
        wpsStart();      
        NextStepWPS = millis()+500;
        FlagWPSActif++;
      }
      else if ( millis() > NextStepWPS)
      {
        NextStepWPS = millis()+500;
        FlagWPSActif++;
        Serial.print(".");
      }
      if ( FlagWPSActif > 200 )
      {
        Serial.printf("\n\nNo answer from WPS, abort wifi connection\n");
        WifiON = false;
        WiFi.disconnect();
        WiFi.mode(WIFI_OFF);          //  Coupe le Wifi
      }
    }
    if ( !FlagWPSActif &&  WiFi.status() == WL_CONNECTED && TimeStampConnectWifi == 0)
    {
      TimeStampConnectWifi = millis();
      Serial.printf("Wifi is now connected, with SSID %s, IP Address: ", WiFi.SSID());
      Serial.println(WiFi.localIP());
      if (!MDNS.begin("Histoire")) {   // Set the hostname to "Histoire.local"
          Serial.println("Error setting up MDNS responder Histoire.local");
          for (int i = 0; i < 100; i++)
          {
            digitalWrite(PIN_LED, i);
            digitalWrite(BUILTIN_LED, i);
            delay(100);
          }
          digitalWrite(PIN_LED, HIGH);
          digitalWrite(BUILTIN_LED, HIGH);
          mDNSStarted = 0;
      }
      else
      {
        Serial.println("mDNS responder started, will answer to Histoire.local");
        mDNSStarted = 1;
        MDNS.addService("http", "tcp", 80);
      }

      if (isConnected()) {
        //  Lance Telnet
        NextChangeLed = millis() + 500;
        NextLEDState = 0;
        ip = WiFi.localIP();
        Serial.println();
        Serial.print("- Telnet: "); Serial.print(ip); Serial.print(":"); Serial.println(portTelnet);
        setupTelnet();
        //  Puis OTA
        // OTA Configiration and Enable OTA
        ArduinoOTA.setHostname("Histoire");
        Serial.println("\nEnabling OTA Feature");
        ArduinoOTA.setPassword("Thierry");
        ArduinoOTA.begin();
        //  Et enfin FTP
        //username, password for ftp.  set ports in ESP32FtpServer.h  (default 21, 50009 for PASV)
        ftpSrv.begin("thierry", "thierry");
        ftpSrv.onFtpConnect(InfoFTPConnect);
        ftpSrv.onFtpDisconnect(InfoFTPDisconnect);
        MDNS.addService("ftp", "tcp", 21);
        MDNS.addService("ftp", "tcp", 22);
        TryConnect = 0;
      }
      TimeStampEndWifi = TimeStampConnectWifi + WifiTimeOut;      //  10 minutes de connexion Wifi avant de couper
    }
    if ( !FlagWPSActif && WiFi.status() != WL_CONNECTED && millis() - LastMessageWifi > 1000 )
    {
      Serial.printf("Try %d, not yet connected, Wifi Status = %d\n", TryConnect, WiFi.status());
      LastMessageWifi = millis();
      TryConnect++;
      if ( (WiFi.status() == WL_NO_SSID_AVAIL || WiFi.status() == WL_CONNECT_FAILED) || TryConnect > 30 )
      {
          Serial.printf("Unable to connect, switching to WPS mode\n");
          FlagWPSActif = 1;
          TimeStampEndWifi = TimeStampStartWifi + 180000;     //  Coupe le Wifi si pas connecté après 3mn
          //wpsStart();
      } 
    }
    if ( millis() > TimeStampEndWifi )
    {
      WifiON = false;
      telnet.println("End of Wifi connection, disconnect all clients");
      Serial.println("End of Wifi connection, disconnect all clients");
      telnet.loop();
      telnet.loop();
      telnet.loop();
      telnet.disconnectClient();
      telnet.loop();
      telnet.loop();
      telnet.loop();
      WiFi.disconnect();
      WiFi.mode(WIFI_OFF);          //  Coupe Wifi à la fin
      NextChangeLed = millis() + 10000000;
      NextLEDState = 1;
    }
    int CurVBUS = analogRead(PIN_VBUS);
    if ( CurVBUS < THRESHOLD_VBUS )
    {
      LastSeenUSBOff = millis();
      isCharging = false;
    }
    else if ( !isCharging && (millis() - LastSeenUSBOff) > 500 )
    {
      isCharging = true;
      if ( !audio.isRunning() )
      {
        audio.connecttoFS(SD, "/Charge.mp3");
      }
    }
    else if ( isCharging && LastBatStatus == STATUS_FULL )
    {
      LastBatStatus = STATUS_OK;
      if ( !audio.isRunning() )
      {
        audio.connecttoFS(SD, "/FinCharge.mp3");
      }
    }
    //  Telnet handle
    telnet.loop();
    // OTA Handle
    ArduinoOTA.handle();
    //  FTP
    ftpSrv.handleFTP();
    if ( FlagCalibration )
    {
      //  Mode calibration TOUCH PAD
      //  attention pas de boucle audio...
      if ( CalibrationOK )
      {
        Pad_OK.startCalibration();
        SaveDataNVS();              //  Sauve valeurs mémorisée
        CalibrationOK = false;
        Serial.println("Calibration OK done, FWD Now");
        CalibrationFWD = true;
      }
      if ( CalibrationFWD )
      {
        Pad_FWD.startCalibration();
        SaveDataNVS();              //  Sauve valeurs mémorisée
        CalibrationFWD = false;
        Serial.println("Calibration FWD done, BACK Now");
        CalibrationBACK = true;
      }
      if ( CalibrationBACK )
      {
        Pad_BACK.startCalibration();
        SaveDataNVS();              //  Sauve valeurs mémorisée
        CalibrationBACK = false;
        FlagCalibration = false;
        Serial.println("Calibration done resume normal operation");
      }
      return;
    }
  }
  else
  {
    digitalWrite(BUILTIN_LED, HIGH);    //  LED off is wifi is off...
    digitalWrite(PIN_LED, HIGH);    //  LED always ON when wifi is off...
    NextChangeLed = millis() + 10000000;
    NextLEDState = 1;
  }    

  int Stat_OK = Pad_OK.getKey();
  if ( Stat_OK ) 
  {
    Serial.printf("Touche OK appuyée, StatOK=%d\n", Stat_OK);
    if ( WifiON )
      telnet.printf("Touche OK appuyée, StatOK=%d\n", Stat_OK);
  }
  int Stat_FWD = Pad_FWD.getKey();
  if ( Stat_FWD ) 
  {
    Serial.printf("Touche OK appuyée, Stat_FWD=%d\n", Stat_FWD);
    if ( WifiON )
    telnet.printf("Touche OK appuyée, Stat_FWD=%d\n", Stat_FWD);
  }
  int Stat_Back = Pad_BACK.getKey();
  if ( Stat_Back ) 
  {
    Serial.printf("Touche OK appuyée, Stat_Back=%d\n", Stat_Back);
    if ( WifiON )
      telnet.printf("Touche OK appuyée, Stat_Back=%d\n", Stat_Back);
  }

  if ( Stat_OK == 2 && !ModeChangeVol )
  {
    Serial.printf("Appui long sur StatOK, enter CHANGE_VOLUME_START\n");
    if ( WifiON )
      telnet.printf("Appui long sur StatOK, enter CHANGE_VOLUME_START\n");
    if ( audio.isRunning() ) audio.stopSong();
    CurrentState = CHANGE_VOLUME_START;
    ModeChangeVol = true;
  }
  if ( millis() > NextCheckBat )
  {
    CheckBatLevel();
  }

  if ( millis() > NextChangeLed )
  {
    digitalWrite(PIN_LED, NextLEDState);
    digitalWrite(BUILTIN_LED, NextLEDState);
    if ( FlagWPSActif > 0 )
      NextChangeLed = millis() + 100;
    else
      NextChangeLed = millis() + 500;
    NextLEDState ^= 1;
  }
  switch (CurrentState)
  {
  case WAIT_FOR_START:
    /* Départ, attente touche pour démarrer */
    if ( Stat_OK || Stat_FWD || Stat_Back )
    {
      Serial.printf("State = WAIT_FOR_START et touche appuyée\n");
      CurrentStory = getNextStory();
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      Serial.printf("Playing Story Title %03d\n", CurrentStory);
    if ( WifiON )
      telnet.printf("Playing Story Title %03d\n", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingTitle = true;
      PlayingStory = false;
      CurrentState = READING_TITLE;
    }
    break;

  case READING_TITLE:
    if ( audio.isRunning() == false )
    {
      CurrentState = WAIT_FOR_OK;
      EndWaitForOK = millis() + 1000;    //  Attente 1s avant validation automatique
      Serial.printf("New State WAIT_FOR_OK, end after 1s\n");
    if ( WifiON )
      telnet.printf("New State WAIT_FOR_OK, end after 1s\n");
    }
    else if ( Stat_OK == 1 )
    {
      CurrentState = READING_STORY;
      PlayingTitle = false;
      SetStoryList();
      Serial.printf("OK pressed duting title, now playing Story body %03d\n", CurrentStory);
    if ( WifiON )
      telnet.printf("OK pressed duting title, now playing Story body %03d\n", CurrentStory);
      snprintf(Buffer, 32, "/Stories/H%03d.mp3", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingStory = true;
      Serial.printf("New State READING_STORY\n");
    if ( WifiON )
      telnet.printf("New State READING_STORY\n");
    }
    else if ( Stat_FWD )
    {
      Serial.printf("State = READING_TITLE et touche FWD appuyée\n");
      if ( WifiON )
        telnet.printf("State = READING_TITLE et touche FWD appuyée\n");
      CurrentStory = getNextStory();
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      Serial.printf("Playing Story Title %03d\n", CurrentStory);
      if ( WifiON )
        telnet.printf("Playing Story Title %03d\n", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingTitle = true;
      PlayingStory = false;
      CurrentState = READING_TITLE;
      Serial.printf("New State READING_TITLE\n");
      if ( WifiON )
        telnet.printf("New State READING_TITLE\n");
    }
    else if ( Stat_Back )
    {
      Serial.printf("State = READING_TITLE et touche BACK appuyée\n");
      if ( WifiON )
        telnet.printf("State = READING_TITLE et touche BACK appuyée\n");
      CurrentStory = getPrevStory();
      if ( CurrentStory > 0 )
      {
        snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
        Serial.printf("Playing Story Title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Playing Story Title %03d\n", CurrentStory);
        audio.connecttoFS(SD, Buffer);
        PlayingTitle = true;
        PlayingStory = false;
        CurrentState = READING_TITLE;
        Serial.printf("New State READING_TITLE\n");
      }
      else
      {
        CurrentStory = IndexRecentStories[0] + 1;
        IndexRecentStories[MAX_BACK_STORY] = CurrentStory - 1;
        Serial.printf("Error, no previous story, set Current story=%d, Play Error sound\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Error, no previous story, set Current story=%d, Play Error sound\n", CurrentStory);
        audio.connecttoFS(SD, "/error.mp3");
        CurrentState = STATE_ERROR;
        PlayingStory = false;
        PlayingTitle = false;
        Serial.printf("New State STATE_ERROR\n");
      }
    }
    break;

  case WAIT_FOR_OK:
    if ( Stat_OK == 1 || millis() > EndWaitForOK )
    {
      CurrentState = READING_STORY;
      PlayingTitle = false;
      Serial.printf("State WAIT_FOR_OK ended, Playing Story body %03d\n", CurrentStory);
      if ( WifiON )
        telnet.printf("State WAIT_FOR_OK ended, Playing Story body %03d\n", CurrentStory);
      SetStoryList();
      snprintf(Buffer, 32, "/Stories/H%03d.mp3", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingStory = true;
      Serial.printf("New State READING_STORY\n");
    }
    else if ( Stat_FWD )
    {
      Serial.printf("State = WAIT_FOR_OK et touche FWD appuyée\n");
      if ( WifiON )
        telnet.printf("State = WAIT_FOR_OK et touche FWD appuyée\n");
      CurrentStory = getNextStory();
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      Serial.printf("State = WAIT_FOR_OK et touche FWD appuyée\n");
      if ( WifiON )
        telnet.printf("State = WAIT_FOR_OK et touche FWD appuyée\n");
      audio.connecttoFS(SD, Buffer);
      PlayingTitle = true;
      PlayingStory = false;
      CurrentState = READING_TITLE;
      Serial.printf("New State READING_TITLE\n");
    }
    else if ( Stat_Back )
    {
      Serial.printf("State = WAIT_FOR_OK et touche BACK appuyée\n");
      if ( WifiON )
        telnet.printf("State = WAIT_FOR_OK et touche BACK appuyée\n");
      CurrentStory = getPrevStory();
      if ( CurrentStory > 0 )
      {
        snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
        Serial.printf("Playing Story Title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Playing Story Title %03d\n", CurrentStory);
        audio.connecttoFS(SD, "/error.mp3");
        PlayingTitle = true;
        PlayingStory = false;
        CurrentState = READING_TITLE;
        Serial.printf("New State READING_TITLE\n");
      }
      else
      {
        IndexRecentStories[MAX_BACK_STORY] = IndexRecentStories[0];
        CurrentStory = IndexRecentStories[0] + 1;
        Serial.printf("Error no previous story, Set Current story=%d, Play Error sound\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Error no previous story, Set Current story=%d, Play Error sound\n", CurrentStory);
        audio.connecttoFS(SD, "/error.mp3");
        CurrentState = STATE_ERROR;
        PlayingStory = false;
        PlayingTitle = false;
        Serial.printf("New State STATE_ERROR\n");
      }
    }
    break;

  case READING_STORY:
    if ( audio.isRunning() == false )
    {
      if ( ModeContinu )
      {
        Serial.printf("Lecture terminée, passe histoire suivanten NextState:WAIT_FOR_NEXT_STORY\n");
        if ( WifiON )
          telnet.printf("Lecture terminée, passe histoire suivanten NextState:WAIT_FOR_NEXT_STORY\n");
        TS_NextStory = millis() + 1000;   //  One second wait before next story
        CurrentState = WAIT_FOR_NEXT_STORY;
      }
      else
      {
        CurrentState = STORY_END;
        Serial.printf("Lecture terminée, new State STORY_END\n");
        if ( WifiON )
          telnet.printf("Lecture terminée, new State STORY_END\n");
      }
    }
    else if ( Stat_OK == 1 )
    {
      //  Passage en pause
      Serial.printf("Pause appuyée suspend lecture\n");
        if ( WifiON )
          telnet.printf("Pause appuyée suspend lecture\n");
      CurrentState = READING_PAUSE;
      audio.pauseResume();
      Serial.printf("New State READING_PAUSE\n");
    }
    else if ( Stat_FWD )
    {
      Serial.printf("State = READING_STORY et touche FWD appuyée\n");
        if ( WifiON )
          telnet.printf("State = READING_STORY et touche FWD appuyée\n");
      CurrentStory = getNextStory();
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      Serial.printf("New story : %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("New story : %03d\n", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingTitle = true;
      PlayingStory = false;
      CurrentState = READING_TITLE;
      Serial.printf("New State READING_TITLE\n");
    }
    else if ( Stat_Back )
    {
      Serial.printf("State = READING_STORY et touche BACK appuyée\n");
        if ( WifiON )
          telnet.printf("State = READING_STORY et touche BACK appuyée\n");
      CurrentStory = getPrevStory();
      if ( CurrentStory > 0 )
      {
        snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
        Serial.printf("Playing Story Title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Playing Story Title %03d\n", CurrentStory);
        audio.connecttoFS(SD, Buffer);
        PlayingTitle = true;
        PlayingStory = false;
        CurrentState = READING_TITLE;
        Serial.printf("New State READING_TITLE\n");
      }
      else
      {
        IndexRecentStories[MAX_BACK_STORY] = IndexRecentStories[0];
        CurrentStory = IndexRecentStories[0] + 1;
        Serial.printf("Error, no previous story set Current story=%d, Play Error sound\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Error, no previous story set Current story=%d, Play Error sound\n", CurrentStory);
        audio.connecttoFS(SD, "/error.mp3");
        CurrentState = STATE_ERROR;
        PlayingStory = false;
        PlayingTitle = false;
        Serial.printf("New State STATE_ERROR\n");
      }
    }
    break;

  case WAIT_FOR_NEXT_STORY:
      if ( millis() > TS_NextStory )
      {
        Serial.println("End of WAIT_FOR_NEXT_STORY, go to next story");
        telnet.println("End of WAIT_FOR_NEXT_STORY, go to next story");
        CurrentStory = getNextStory();
        snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
        Serial.printf("Playing Story Title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Playing Story Title %03d\n", CurrentStory);
        audio.connecttoFS(SD, Buffer);
        PlayingTitle = true;
        PlayingStory = false;
        CurrentState = READING_TITLE;
      }
      break;
  case READING_PAUSE:
    if ( Stat_OK == 1 )
    {      //  Passage en Play
      Serial.printf("State READING_PAUSE --> Play appuyée, reprend lecture\n");
        if ( WifiON )
          telnet.printf("State READING_PAUSE --> Play appuyée, reprend lecture\n");
      CurrentState = READING_STORY;
      audio.pauseResume();      
      Serial.printf("New State READING_STORY\n");
    }
    else if ( Stat_Back )
    {
      audio.stopSong();     //  Arête lecture pour repartir au début
      CurrentState = READING_TITLE;
      PlayingTitle = true;
      Serial.printf("PAUSE et back appuyé, Restarting playing Story title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("PAUSE et back appuyé, Restarting playing Story title %03d\n", CurrentStory);
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingStory = false;
      Serial.printf("New State READING_TITLE\n");
    }
    else if ( Stat_FWD )
    {
      audio.stopSong();     //  Arête lecture pour aller au suivant
      Serial.printf("State = READING_PAUSE et touche FWD appuyée\n");
        if ( WifiON )
          telnet.printf("State = READING_PAUSE et touche FWD appuyée\n");
      CurrentStory = getNextStory();
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      Serial.printf("Playing Story Title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Playing Story Title %03d\n", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingTitle = true;
      PlayingStory = false;
      CurrentState = READING_TITLE;
      Serial.printf("New State READING_TITLE\n");
    }
    break;

  case STORY_END:
    if ( Stat_Back )
    {
      CurrentState = READING_TITLE;
      PlayingTitle = true;
      Serial.printf("State STORY_END at BACK appuyé, Retarting playing Story title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("State STORY_END at BACK appuyé, Retarting playing Story title %03d\n", CurrentStory);
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingStory = false;
      Serial.printf("New State READING_TITLE\n");
    }
    else if ( Stat_OK  == 1 )
    {
      Serial.printf("State = STORY_END et touche OK appuyée\n");
        if ( WifiON )
          telnet.printf("State = STORY_END et touche OK appuyée\n");
      CurrentStory = getNextStory();
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      Serial.printf("Playing Story Title %03d\n", CurrentStory);
        if ( WifiON )
          telnet.printf("Playing Story Title %03d\n", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingTitle = true;
      PlayingStory = false;
      CurrentState = READING_TITLE;
      Serial.printf("New State READING_TITLE\n");
    }
    else if ( Stat_FWD  )
    {
      Serial.printf("State = STORY_END et touche FWD appuyée\n");
      if ( WifiON )
        telnet.printf("State = STORY_END et touche FWD appuyée\n");
      CurrentStory = getNextStory();
      snprintf(Buffer, 32, "/Stories/T%03d.mp3", CurrentStory);
      Serial.printf("Playing Story Title %03d\n", CurrentStory);
      if ( WifiON )
        telnet.printf("Playing Story Title %03d\n", CurrentStory);
      audio.connecttoFS(SD, Buffer);
      PlayingTitle = true;
      PlayingStory = false;
      CurrentState = READING_TITLE;
      Serial.printf("New State READING_TITLE\n");
    }
    break;

  case STATE_ERROR:
    if ( audio.isRunning() == false )
    {
      Serial.printf("Fin son erreur\n");
        if ( WifiON )
          telnet.printf("Fin son erreur\n");
      CurrentState = STORY_END;
      Serial.printf("New State STORY_END\n");
        if ( WifiON )
          telnet.printf("New State STORY_END\n");
    }
  case CHANGE_VOLUME_START:
      audio.connecttoFS(SD, "/Change.mp3");
      CurrentState = CHANGE_VOLUME_1;
      break;

  case CHANGE_VOLUME_1:
    if ( audio.isRunning() == false )
      CurrentState = CHANGE_VOLUME_WAIT;
    break;
  
  case CHANGE_VOLUME_WAIT:
    if ( Stat_Back )
    {
      if ( audio.isRunning() ) audio.stopSong();
      if ( CurrentVolume > 0 )  CurrentVolume--;
      Serial.printf("Nouveau volume %d\n", CurrentVolume);
      audio.setVolume(CurrentVolume);
      audio.connecttoFS(SD, "/NewVol.mp3");
    }
    else if ( Stat_FWD )
    {
      if ( audio.isRunning() ) audio.stopSong();
      if ( CurrentVolume < MAX_VOLUME )  CurrentVolume++;
      Serial.printf("Nouveau volume %d\n", CurrentVolume);
      audio.setVolume(CurrentVolume);
      audio.connecttoFS(SD, "/NewVol.mp3");
    }
    if ( Pad_OK.StillPressed() == 0 )
    {
      if ( audio.isRunning() ) audio.stopSong();
      CurrentState = STORY_END;     //  Fin
      ModeChangeVol = false;
      SaveDataNVS();
    }
  default:
    break;
  }

  audio.loop();

}

// optional
void audio_info(const char *info){
    //Serial.print("info        "); Serial.println(info);
}
void audio_id3data(const char *info){  //id3 metadata
    //Serial.print("id3data     ");Serial.println(info);
}
void audio_eof_mp3(const char *info){  //end of file
   //Serial.print("eof_mp3     ");Serial.println(info);
}
