//  Header file for Touch PAD
#include "Arduino.h"
#define TELNET_DEBUG 1

#if TELNET_DEBUG > 0
extern ESPTelnet telnet;
#endif

#define RESET_STAT_PERIOD   1000
#define MAX_CHANGE          10000
class Touch
{
private:
    uint8_t m_pinNum;
    uint8_t TransientState;
    uint8_t ValidatedState;
    uint16_t m_Threshold;
    uint16_t MinValuePAD;
    uint16_t MaxValuePAD;

    uint32_t Debounce_HL;
    uint32_t Debounce_LH;

    uint32_t DelayLong;

    uint32_t SumValuePAD;

    uint32_t LastChange;
    uint32_t TS_CurVal;

    uint32_t nChange;
    uint32_t nCall;
    
    uint32_t TimeSpent;

    uint32_t PeriodStat;
    uint32_t nCallStat;

    uint8_t Lastkey;
    bool    ModeDebug;

    char m_Name[32];

public:
    Touch(unsigned char pinNumber, unsigned int Threshold );
    ~Touch();

    int rawDetect();
    int getKey();
    int StillPressed();
    void setThreshold(uint16_t Threshold);
    void setAutoThreshold();
    uint16_t getThreshold();
    void setDebounce(uint32_t lh, uint32_t hl);
    void setName(const char *name);
    void setStatPeriod(uint32_t Period);
    void setDelayLong(uint32_t Delay);
    void setModeDebug(bool FlagDebug) { ModeDebug = FlagDebug;}
    void startCalibration();
};

Touch::Touch(unsigned char pinNumber, unsigned int Threshold)
{
    m_pinNum = pinNumber;
    m_Threshold = Threshold;
    ValidatedState = 0;
    TransientState = 0;
    MinValuePAD = 65535;
    MaxValuePAD = 0;
    SumValuePAD = 0;
    nChange = 0;
    nCall = 0;
    TimeSpent = 0;
    ModeDebug = false;
    nCallStat = 0;
    Lastkey = 0;
    Debounce_HL = 300000;       //  300 ms pour sélection 
    Debounce_LH = 100000;       //  100 ms pour désélection 
    LastChange = micros();
    TS_CurVal = LastChange;
    DelayLong = 20000000;       //  20s pour appui long par défaut
    PeriodStat = RESET_STAT_PERIOD;
    snprintf(m_Name, 16, "%d", pinNumber);
}

Touch::~Touch()
{
}

//  Fixe le seuils de détection

void Touch::setThreshold(uint16_t Threshold)
{
    m_Threshold = Threshold;
}


//  Essaie de fixer le seuil automatiquement
//  Lecture 100 fois du seuil sans être appuyé, prend la moyenne et multiplie par 0.95
//  Cette fonction ne doit pas être appelée en fonctionnement "normal" mais pendant le setup

void Touch::setAutoThreshold()
{
uint32_t ValSomme = 0;

    for ( int i = 0; i < 100; i++ )
    {
        int16_t ValTouch = touchRead(m_pinNum);
        ValSomme += ValTouch;
        delay(2);
    }
    uint32_t Average = ValSomme / 100;
    uint16_t NewThresold = (Average * 0.95);
    Serial.printf("setAutoThreshold touche %s(%d), Moyenne=%d, Nouveau seuil=%d\n", m_Name, m_pinNum, Average, NewThresold);
    m_Threshold = NewThresold;
}

//  Calibration assistée de la touche correspondante
//  Cette fonction ne doit pas être appelée en fonctionnement "normal" pas de boucle audio, seulement Telnet

void Touch::startCalibration()
{
uint32_t ValSommeOff = 0;
uint16_t ValMinOff = 32767;
uint16_t ValMaxOff = -32768;
uint32_t ValSommeOn = 0;
uint16_t ValMinOn = 32767;
uint16_t ValMaxOn = -32768;

    //  Seulement avec Telnet...
    Serial.printf("N'appuyez PAS sur la touche %s\n", m_Name);
#if TELNET_DEBUG > 0
    telnet.printf("N'appuyez PAS sur la touche %s\n", m_Name);
    telnet.loop();
#endif
    //  Première phase appel 1000 fois sans toucher
    for ( int i = 0; i < 1000; i++ )
    {
        int16_t ValTouch = touchRead(m_pinNum);
        if ( ValTouch < ValMinOff ) ValMinOff = ValTouch;
        if ( ValTouch > ValMaxOff ) ValMaxOff = ValTouch;
        ValSommeOff += ValTouch;
        delay(5);
    }
    uint32_t AverageOff = ValSommeOff / 1000;
    Serial.printf("1000 appels touchRead(%s), Min=%d, Max=%d, Avg=%d\n", m_Name, ValMinOff, ValMaxOff, AverageOff);
#if TELNET_DEBUG > 0
    telnet.printf("1000 appels touchRead(%s), Min=%d, Max=%d, Avg=%d\n", m_Name, ValMinOff, ValMaxOff, AverageOff);
    telnet.loop();
    telnet.printf("Appuyez sur la touche %s\n", m_Name);
    telnet.loop();
#endif
    //  Seconde phase appel 1000 fois sur touche appuyée

    Serial.printf("Appuyez sur la touche %s\n", m_Name);
    //  Attend environ 1s avant de passer à l'étape suivante
    for ( int i = 0; i < 1000; i++ )
    {
        delay(1);
#if TELNET_DEBUG > 0
        telnet.loop();
#endif
    }
    for ( int i = 0; i < 1000; i++ )
    {
        int16_t ValTouch = touchRead(m_pinNum);
        if ( ValTouch < ValMinOn ) ValMinOn = ValTouch;
        if ( ValTouch > ValMaxOn ) ValMaxOn = ValTouch;
        ValSommeOn += ValTouch;
        delay(5);
    }
    uint32_t AverageOn = ValSommeOn / 1000;
    Serial.printf("1000 appels touchRead(%s), Min=%d, Max=%d, Avg=%d\n", m_Name, ValMinOn, ValMaxOn, AverageOn);
#if TELNET_DEBUG > 0
    telnet.printf("1000 appels touchRead(%s), Min=%d, Max=%d, Avg=%d\n", m_Name, ValMinOn, ValMaxOn, AverageOn);
    telnet.loop();
    telnet.printf("Appuyez sur la touche %s\n", m_Name);
    telnet.loop();
#endif
    //  Calcul du seuil : moyenne entre les 2 valeurs
    uint16_t NewThresold = (AverageOn + AverageOff) / 2;     //  Première valeur seuil
    setThreshold(NewThresold);
    Serial.printf("Valeur seuil calculée pour la touche %s : %d\n", m_Name, NewThresold);
    Serial.println("Test avec les valeurs calculées");
#if TELNET_DEBUG > 0
    telnet.printf("Valeur seuil calculée pour la touche %s : %d\n", m_Name, NewThresold);
    telnet.println("Test avec les valeurs calculées");
    telnet.loop();
#endif

    //  Attend environ 1s avant de passer à l'étape suivante
    for ( int i = 0; i < 1000; i++ )
    {
        delay(1);
#if TELNET_DEBUG > 0
        telnet.loop();
#endif
    }
    //  Dernière phase TEST valeur du seuil pendant 10s
    uint32_t EndTest = millis() + 10000;
    while ( millis() < EndTest )
    {
        int State = getKey();
        if ( State )
        {
//            Serial.printf("Touche %s appuyée\n", m_Name);
#if TELNET_DEBUG > 0
            telnet.printf("Touche %s appuyée\n", m_Name);
            telnet.loop();
#endif
        }
    }
    //  Attend environ 1s avant de passer à l'étape suivante
    for ( int i = 0; i < 1000; i++ )
    {
        delay(1);
#if TELNET_DEBUG > 0
        telnet.loop();
#endif
    }
    Serial.printf("Valeur seuil calculée pour la touche %s : %d\n", m_Name, NewThresold);
    Serial.printf("Fin de la calibration pour la touche %s\n", m_Name);
#if TELNET_DEBUG > 0
    telnet.printf("Valeur seuil calculée pour la touche %s : %d\n", m_Name, NewThresold);
    telnet.printf("Fin de la calibration pour la touche %s\n", m_Name);
    telnet.loop();
    telnet.loop();
    telnet.loop();
#endif
}

uint16_t Touch::getThreshold()
{
    return m_Threshold;
}

//  Fixe les temps de Debounce

void Touch::setDebounce(uint32_t lh, uint32_t hl)
{
    Debounce_HL = hl;
    Debounce_LH = lh;
}


//  Fixe le nom du PAD pour faciliter le Debug

void Touch::setName(const char *Name)
{
    strncpy(m_Name, Name, 31);
    m_Name[31] = 0;
}

//  Fixe la periode pour les statistiques
//  Si 0, pas de stats...

void Touch::setStatPeriod(uint32_t Period)
{
    if ( ModeDebug) Serial.printf("Changing Stat Period, After %d calls for Pad %s, Min=%d, Max=%d, Avg=%d, TimeSpent=%d µs/call\n", 
        nCall, m_Name, MinValuePAD, MaxValuePAD, SumValuePAD/PeriodStat, TimeSpent/PeriodStat);
    PeriodStat = Period;
    nCallStat = 0;
}

void Touch::setDelayLong(uint32_t Delay)
{
    DelayLong = Delay;
}

//  Lecture de l'état des touches
//  Retourne 1 si la touche est "appuyée"

int Touch::rawDetect()
{
    uint32_t CurTS = micros();
    int16_t ValTouch = touchRead(m_pinNum);
    nCall++;
    nCallStat++;
    if ( ValTouch < MinValuePAD ) MinValuePAD = ValTouch;
    if ( ValTouch > MaxValuePAD ) MaxValuePAD = ValTouch;
    SumValuePAD += ValTouch;
    if ( PeriodStat > 0 && nCallStat >= PeriodStat)
    {
        if ( ModeDebug ) 
        {
            Serial.printf("After %d calls for Pad %s, Min=%d, Max=%d, Avg=%d, Last=%d, TimeSpent=%d µs/call\n", 
                nCall, m_Name, MinValuePAD, MaxValuePAD, SumValuePAD/PeriodStat, ValTouch, TimeSpent/PeriodStat);
        }
        MinValuePAD = 65535;
        MaxValuePAD = 0;
        SumValuePAD = 0;
        TimeSpent = 0;
        nCallStat = 0;
    }

    bool CurState = (ValTouch <= m_Threshold);

    //  Check if state has changed
    if ( CurState == TransientState )
    {
        //  Same state
        if ( !ValidatedState &&  TransientState )    //  Pressed but not yet validated ?
        {
            if ( (CurTS - LastChange) > Debounce_HL )
            {
                if ( ModeDebug ) Serial.printf("PAD %s is now detected at TS=%lu, after nChange=%d, LastChange was %lu\n",
                    m_Name, CurTS, nChange, LastChange);
#if TELNET_DEBUG > 0
                if ( ModeDebug ) telnet.printf("PAD %s is now detected at TS=%lu, after nChange=%d, LastChange was %lu\n",
                    m_Name, CurTS, nChange, LastChange);
                telnet.loop();
#endif
                ValidatedState = 1;        //  Set state OK
                TS_CurVal = CurTS;
                nChange = 0;               //  Reset changed counter
            }
        }
        else if ( ValidatedState &&  !TransientState )    //  Not Pressed anymore but not yet validated
        {
            if ( (CurTS - LastChange) > Debounce_LH )
            {
                if ( ModeDebug ) Serial.printf("PAD %s is now NOT detected at TS=%lu, after nChange=%d, LastChange was %lu\n",
                    m_Name, CurTS, nChange, LastChange);
#if TELNET_DEBUG > 0
                if ( ModeDebug ) telnet.printf("PAD %s is now NOT detected at TS=%lu, after nChange=%d, LastChange was %lu\n",
                    m_Name, CurTS, nChange, LastChange);
                telnet.loop();
#endif
                    ValidatedState = 0;        //  Set state NOK
                TS_CurVal = CurTS;
                nChange = 0;               //  Reset changed counter
            }
        }
        else if ( !TransientState )
        {
        //  In this case (no touch) change TS to avoid overflow after long period without touches
            LastChange = CurTS;
        }
  }
  else
  {
    //  State has changed
    LastChange = CurTS;             //  Set Time stamp
    TransientState = CurState;      //  And state
    nChange++;
    if ( nChange > MAX_CHANGE )      // Too many changes, reset state
    {
      if ( ModeDebug ) Serial.printf("Too many changes, reset State for Touch %s\n", m_Name);
      nChange = 0;
      ValidatedState = 0;
      TransientState = 0;
      TS_CurVal = CurTS;
    }
  }
  TimeSpent += micros() - CurTS;
  return ValidatedState;
}

//  Retourne 1 quand la touche est appuyée la première fois
//  Retourne 2 en cas d'appui long
int Touch::getKey()
{
    int state = rawDetect();
    uint32_t curTS = micros();

    if ( state )
    {
        //  Touche appuyée
        if ( (curTS - TS_CurVal) > DelayLong )
        {
            Serial.printf("For Touch %s at TS=%lu, State is ON since TS=%lu (%lu ms)\n", m_Name, curTS, TS_CurVal, (curTS - TS_CurVal)/1000);
#if TELNET_DEBUG > 0
            telnet.printf("For Touch %s at TS=%lu, State is ON since TS=%lu (%lu ms)\n", m_Name, curTS, TS_CurVal, (curTS - TS_CurVal)/1000);
            telnet.loop();
            telnet.loop();
            telnet.loop();
#endif
            TS_CurVal = curTS;              //  Change valeur pour éviter les répétitions
            Lastkey = 1;
            return 2;
        }
        if ( !Lastkey )
        {
            Lastkey = 1;
            return 1;                       //  Seulement au changement d'état
        }
    }
    else
    {
        Lastkey = 0;                        // Touch not pressed anymore
    }
    return 0;
}


int Touch::StillPressed()
{
    return rawDetect();
}
